<?php
/**
 * ManuAdminMod
 *
 * This is a powerful and platform independent software for administrating game servers of various kinds.
 * If you need help with installing or using this software, please visit our website at: www.ManuAdminMod.de
 * If you want to obtain additional permissions extending the license, contact us at: Webmaster@ManuAdminMod.de
 *
 * @copyright www.ManuAdminMod.de
 * @license http://www.creativecommons.org/licenses/by-nc-nd/4.0/ Creative Commons BY-NC-ND 4.0
 * @version 1.0.0-Beta+7
 * */

namespace MAM\Daemon\Engine\Frostbite3;

use MAM\Daemon\Core\Registry;

/**
 * Class to handle all log actions
 *
 * This class handles all log actions
 *
 * @author Master of Little <masteroflittle@manuadminmod.de>
 */
class ActionHandler {
    /**
     * Class for mam related stuff
     *
     * @var \MAM\Daemon\Engine\Frostbite3\Mod
     */
    private $mod;
    /**
     * Class for Remote Commands
     *
     * @var \MAM\Daemon\Engine\Frostbite3\Rcon\RconCommands
     */
    private $rcon;
    /**
     * Class for player related stuff
     *
     * @var \MAM\Daemon\Engine\Frostbite3\Player
     */
    private $players;
    /**
     * Class for Logging
     *
     * @var \MAM\Daemon\Libraries\Logging\LogHandler
     */
    private $logging;
    /**
     * Class for Frostbite tools
     *
     * @var \MAM\Daemon\Engine\Frostbite3\Tools
     */
    private $tools;
    /**
     * Array for Failed Join Attempts to avoid warning at leaving the server
     * @var array
     */
    private $fail_join_attempt = array();

    /**
     * Constructor for ActionHandler class
     */
    public function __construct() {
        $registry = Registry::getInstance();
        $this->mod = $registry->mod;
        $this->players = &$registry->players;
        $this->rcon = $registry->rcon;
        $this->logging = $registry->logging;
        $this->tools = $registry->tools;
    }

    /**
     * Returns a string if object is printed
     *
     * Notice: all plugins which where moved to core needs a __tostring function
     * to get it working with the permissions system. Therefore the return of __tostring must me the
     * namespace like deamon, punishment, permissions ...
     *
     * @see http://php.net/manual/en/language.oop5.magic.php#object.tostring
     * @return string
     */
    public function __toString() {
        return "Frostbite Actionhandler Class";
    }

    /**
     * Log action: player joins the server.
     * @param string $name
     */
    public function Join($name) {
        if ($this->mod->findPlayerGuidByName($name)) {
            return;
        }
        $player_list = $this->rcon->rcon(array("admin.listPlayers", "player", $name));
        if (!isset($player_list[14])) {
            $this->fail_join_attempt[] = $name;
            return;
        }
        $guid = strtolower($player_list[14]);
        $this->players[$guid] = new Player($guid, 99, $name);

        $search = array(
            '<NAME>',
            '<PID>',
            '<GUID>',
            '<PBGUID>',
            '<GROUP>'
        );
        $replace = array(
            $this->players[$guid]->getName(),
            $this->players[$guid]->getPid(),
            $guid,
            substr($guid, -8),
            $this->players[$guid]->getGroup(),
        );
        $string = str_replace($search, $replace, "<NAME> is joining, GUID: <GUID>");
        if (trim($string) != "") {
            $this->logging->info($string);
        }
        $this->players[$guid]->__set("joining", true);
    }

    /**
     * Log action: player joins team.
     * @param string $name
     * @param string $team
     */
    public function JoinTeam($name, $team) {

        if (!$this->mod->findPlayerGuidByName($name)) {
            $this->logging->warning("JOINTEAM by player '$name' who is actually not on the server");
            return;
        }

        $guid = $this->mod->findPlayerGuidByName($name);
        $pid = $this->players[$guid]->getPID();

        if ($this->players[$guid]->__get("joining") == true) {
            $this->players[$guid]->__set("joining", false);
            $this->players[$guid]->update($pid, $name);
            $search = array(
                '<NAME>',
                '<PID>',
                '<GUID>',
                '<PBGUID>',
                '<IP>',
                '<GROUP>'
            );
            $replace = array(
                $this->players[$guid]->getName(),
                $this->players[$guid]->getPid(),
                $guid,
                substr($guid, -8),
                $this->players[$guid]->getIP(),
                $this->players[$guid]->getGroup(),
            );
            $string = str_replace($search, $replace, "<GROUP> <NAME> (<PID>) has joined. Guid:<GUID>");
            if (trim($string) != "") {
                $this->logging->info($string);
            }
            $this->mod->triggerEvent("playerJoined", $guid);
        }
        $this->players[$guid]->update($pid, $name);
        $oldteam = $this->players[$guid]->getTeam();
        $this->players[$guid]->setTeam($team);

        $this->mod->triggerEvent("playerJoinTeam", $guid, $oldteam);
    }

    /**
     * Log action: player joins squad.
     * @param string $name
     * @param string $squad
     */
    public function JoinSquad($name, $squad) {
        if (!$this->mod->findPlayerGuidByName($name)) {
            if ((int) $squad == 0) {
                return;
            }
            $this->logging->warning("JOINSQUAD by player '$name' who is actually not on the server");
            return;
        }
        $guid = $this->mod->findPlayerGuidByName($name);
        $pid = $this->players[$guid]->getPID();
        $this->players[$guid]->update($pid, $name);
        $oldsquad = $this->players[$guid]->getSquad();
        $this->players[$guid]->setSquad($squad);
        $this->mod->triggerEvent("playerJoinSquad", $guid, $oldsquad);
    }

    /**
     * Log action: player leaves the server.
     *
     * @param string $name
     * @return void
     */
    public function Quit($name) {
        if (in_array($name, $this->fail_join_attempt)) {
            unset($this->fail_join_attempt);
            return;
        }

        if (!$this->mod->findPlayerGuidByName($name)) {
            $this->logging->warning("QUIT by player '$name' who is actually not on the server");
            return;
        }
        $guid = $this->mod->findPlayerGuidByName($name);

        if ($this->players[$guid]->__get("joining") == true) { /* When he is joining no stats etc... */
            $this->logging->info("Player '$name' quit, PID: " . $this->players[$guid]->getPID() . ", GUID: $guid");
            unset($this->players[$guid]);
            return;
        }
        $this->mod->triggerEvent("playerQuit", $guid);
        $this->logging->info("Player '$name' quit, PID: " . $this->players[$guid]->getPID() . ", GUID: $guid");
        $this->players[$guid]->cleanup();
        unset($this->players[$guid]);
    }

    /**
     * Log action: player says something, if message starts with command prefix, the command will be executed.
     * @param string $name
     * @param string $msg
     */
    public function Say($name, $msg) {
        //0:GUID
        //1:PID
        //2:Name
        //3:Msg
        if (!$this->mod->findPlayerGuidByName($name)) {
            $this->logging->warning("SAY by player '$name' who is actually not on the server");
            return;
        }
        $guid = $this->mod->findPlayerGuidByName($name);

        if (empty($msg)) { /* non printable chars */
            return;
        }

        $executed = false;
        //var_dump($parts[3]);
        if ($msg[0] == "/" && "/" != $this->mod->getCV("daemon", "cmdprefix")) {
            $msg = substr($msg, 1);
        }

        if ($msg[0] == $this->mod->getCV("daemon", "cmdprefix")) {
            $this->mod->executeCommand($msg, $guid, $executed);
        }

        $this->mod->triggerEvent("playerSay", array($guid, $msg, $executed));
    }

    /**
     * Log action: next map / map restart
     * @param array $dvars
     */
    public function NextMap($dvars) {
        $this->logging->info("Next map / map restart");

        $this->gametype = $dvars["g_gametype"];
        $this->mod->lastmaps[] = $dvars["mapname"];

        $this->logging->info("Current map: " . $dvars["mapname"] . " (" . $dvars["g_gametype"] . ")");

        foreach ($this->players as $key => $value) {
            $this->players[$key]->resetTeam();
            $this->players[$key]->resetSquad();
        }

        //$this->mod->updateTeamNames();
        /**
         * @todo Check execute time, took to long. Do we need that in BF4?
         */
        /*
          //load map specific config files
          if (is_file($this->mod->getConfigDir() . "/maps/default.json")) {
          $this->mod->readConfigFile("maps/default.json");
          }
          if (is_file($this->mod->getConfigDir() . "/maps/" . $dvars["mapname"] . ".json")) {
          $this->mod->readConfigFile("maps/" . $dvars["mapname"] . ".json");
          }

          if (is_readable($this->mod->getConfigDir() . "/maps/default.rcon")) {
          $this->logging->info("Executing: maps/default.rcon");
          $cmds = file($this->mod->getConfigDir() . "/maps/default.rcon");
          foreach ($cmds as $cmd) {
          $cmd = preg_replace('|(//).*|i', '', $cmd);
          if (!empty($cmd)) {
          $this->rcon->rconRcon(trim($cmd));
          }
          }
          }

          if (is_readable($this->mod->getConfigDir() . "/maps/" . $dvars["mapname"] . ".rcon")) {
          $this->logging->info("Executing: maps/" . $dvars["mapname"] . ".rcon");
          $cmds = file($this->mod->getConfigDir() . "/maps/" . $dvars["mapname"] . ".rcon");
          foreach ($cmds as $cmd) {
          $cmd = preg_replace('|(//).*|i', '', $cmd);
          if (!empty($cmd)) {
          $this->rcon->rconRcon(trim($cmd));
          }
          }
          }
         */
        $this->mod->triggerEvent("nextMap", $dvars);
    }

    /**
     * Runs when a server restart ist detected.
     */
    public function ServerRestart() {
        $this->logging->info("Detected a server restart");
        $this->players = array();
        $this->mod->triggerEvent("serverRestart");
    }

    /**
     * Log action: Win
     * @param int $team
     */
    public function ResultBF4($team) {
        $this->mod->triggerEvent("roundover", $team);
        $this->logging->info("Round over winner: $team (" . $this->mod->getLongTeamName($team) . ")");
    }

    /**
     * Log action: Punkbuster logline
     * @param array $line
     */
    public function punkBuster($line) {
        $this->tools->punkBusterPlayer(false, $line);
        $this->mod->triggerEvent("punkBuster", $line);
    }

    /**
     * Log action: player kills another player (or himself ;-)
     * @param $killer_name
     * @param $victim_name
     * @param $weapon
     * @param $bodypart
     */
    public function Kill($killer_name, $victim_name, $weapon, $bodypart) {

        if ($killer_name === '') {
            return;
        }

        if (!$this->mod->findPlayerGuidByName($killer_name)) {
            $this->logging->warning("KILL by player '$killer_name' who is actually not on the server");
            return;
        }
        if (!$this->mod->findPlayerGuidByName($victim_name)) {
            $this->logging->warning("DEATH by player '$victim_name' who is actually not on the server");
            return;
        }
        $killer_guid = $this->mod->findPlayerGuidByName($killer_name);
        $victim_guid = $this->mod->findPlayerGuidByName($victim_name);

        $victim_team = $this->players[$victim_guid]->getTeam();
        $killer_team = $this->players[$killer_guid]->getTeam();

        if ($killer_name == $victim_name) {
            $type = "selfkill";
        } elseif ($killer_team != "none" && $victim_team != "none" && $killer_team === $victim_team) {
            $type = "teamkill";
        } else {
            $type = "kill";
        }
        if ($type == "teamkill" && $weapon == "U_Medkit") {
            return;
        }
        //  echo "$type: $killer_name(". $this->players[$killer_guid]->getTeam()."), $victim_name(". $this->players[$victim_guid]->getTeam().")  $bodypart\n";
        $this->players[$victim_guid]->death($type, $killer_guid, array($weapon, false), 100, $bodypart);
        $this->players[$killer_guid]->kills($type, $victim_guid, array($weapon, false), 100, $bodypart);
    }
}
