<?php
/**
 * ManuAdminMod
 *
 * This is a powerful and platform independent software for administrating game servers of various kinds.
 * If you need help with installing or using this software, please visit our website at: www.ManuAdminMod.de
 * If you want to obtain additional permissions extending the license, contact us at: Webmaster@ManuAdminMod.de
 *
 * @copyright www.ManuAdminMod.de
 * @license http://www.creativecommons.org/licenses/by-nc-nd/4.0/ Creative Commons BY-NC-ND 4.0
 * @version 1.0.0-Beta+7
 * */

namespace MAM\Daemon\Libraries\Database;

use \PDO;

/**
 * Database Class
 *
 * This class is a powerfull class to handle your database connection
 * through PDO as database driver it#s possible to use nearly every database
 * This class also supports transactions with rollbacks
 *
 * @author Mirko911 <mirko911@manuadminmod.de>
 */
class Database {
    /**
     * Database handler
     * @var PDO
     */
    protected $dbh;
    /**
     * Value of the last inserted id
     *
     * @var int
     */
    protected $last_inserted_id;
    /**
     * How to order the get sql query
     *
     * @var string
     */
    protected $order;
    /**
     * How to limit the get sql query
     * @var string
     */
    protected $limit;
    /**
     *
     * PDO Statement
     *
     * @var PDOStatement
     */
    protected $sth;
    /**
     * The database prefix
     *
     * For example mam_
     *
     * @var string
     */
    protected $prefix;
    /**
     * Fetch type for database
     * @var type
     */
    protected $fetch_type;
    /**
     * Saves the parameters for PDO
     *
     * @var array
     */
    protected $options = array(
        PDO::ATTR_PERSISTENT => true,
        PDO::ATTR_ERRMODE => PDO::ERRMODE_WARNING,
        PDO::ATTR_DEFAULT_FETCH_MODE => PDO::FETCH_OBJ,
    );

    /**
     * Set Parameters for PDO
     *
     * @link http://is1.php.net/manual/en/pdo.setattribute.php List of Paramters
     *
     * @param array $options Key => $value where both are PDO Constatns
     */
    public function setParameters($options) {
        foreach ($options as $key => $value) {
            $this->options[$key] = $value;
        }
    }

    /**
     * Set Prefix for Tables
     *
     * Every Table gets this pfx | except you use the query/exec function
     *
     * @param string $prefix
     */
    public function setPrefix($prefix) {
        $this->prefix = $prefix;
    }

    /**
     * Sets the fetch type for 1 query
     *
     * @see http://www.php.net/manual/de/pdostatement.fetch.php
     * @param int $type Pdo fetch type
     */
    public function setFetchType($type) {
        $this->fetch_type = $type;
    }

    /**
     * Returns the Prefix
     *
     * @return string
     */
    public function getPrefix() {
        return $this->prefix;
    }

    /**
     * Constrcutor function
     *
     * @link http://www.php.net/manual/de/pdo.drivers.php List of drivers you can use
     * @param   string    $dsn      Data Source Name
     * @param   string    $user     Username [optional]
     * @param   string    $password Password [optional]
     */
    public function __construct($dsn, $user = '', $password = '') {
        $this->connect($dsn, $user, $password);
    }

    /**
     * Desturctor
     */
    public function __destruct() {
        $this->dbh = null;
    }

    /**
     * Returns a string if object is printed
     *
     * Notice: all plugins which where moved to core needs a __tostring function
     * to get it working with the permissions system. Therefore the return of __tostring must me the
     * namespace like deamon, punishment, permissions ...
     *
     * @see http://php.net/manual/en/language.oop5.magic.php#object.tostring
     * @return string
     */
    public function __toString() {
        return 'Database Class';
    }

    /**
     * Connects to Database
     *
     * @param string $dsn Data Source Name
     * @param string $user Username if needed, otherwise it's an empty string
     * @param string $password Password if needed, otherwise it's an empty string
     */
    public function connect($dsn, $user, $password) {
        $this->dbh = null;
        try {
            $this->dbh = new PDO($dsn, $user, $password, $this->options);
        } catch (\PDOException $ex) {
            \MAM\Daemon\Core\Registry::getInstance()->logging->error('Database Class: ' . $ex);
        }
    }

    /**
     * Binds a value to a predefined parameter
     *
     * With this function it's possible to execute a query multiple
     * times with different values
     *
     * @param   string  $name   Parameter Name
     * @param   mixed   $value  Parameter Value
     * @param   int     $type   PDO Param Type [optional]
     */
    public function bindValue($name, $value, $type = NULL) {
        if (is_null($type)) {
            $type = $this->getPDOType($value);
        }
        $this->sth->bindValue($name, $value, $type);
    }

    /**
     * Adds a specific order for the get function
     *
     * you can use this function multiple times to set multiple orders
     *
     * @param   string  $column Name of the column
     * @param   string  $type ASC or DESC
     */
    public function addOrder($column, $type = '') {
        if (empty($this->order)) {
            $this->order = ' ORDER by ';
        }
        $this->order .= '`' . $column . '`' . ' ' . $type;
    }

    /**
     * Removes the order, so it can't affect the next query
     */
    public function clearOrder() {
        $this->order = '';
    }

    /**
     * Sets a limit for the get function
     *
     * @param   int $startvalue ID to start from
     * @param   int $count      How much results to you want to get?
     */
    public function setLimit($startvalue, $count) {
        $this->limit = ' LIMIT ' . $startvalue . ',' . $count;
    }

    /**
     * Removes the limit, so it can't affect the next query
     */
    public function clearLimit() {
        $this->limit = '';
    }

    /**
     * Gets Values from the database
     *
     * If you pass an array as input you can directly affect the where clause
     * Example:
     * $db->get("mam_users", array("id",5)); => "SELECT * FROM mam_users WHERE id = 5
     *
     * If you only need some fields you can pass a numeric array with the fields you want
     * Example:
     * $db->get("mam_users", array("id",5), array("guid", "lastseen")); => "SELECT guid,lastseen FROM mam_users WHERE id = 5
     *
     * If you pass a string as input, you'll pass a sql query
     * Example:
     * $db->get("mam_users", "SELECT * FROM <TABLE> WHERE guid = :guid, array(":guid" => $guid));
     * For SQL Querys it's possible to use the placeholders <TABLE> for table with prefix or <PFX> for prefix
     *
     * The functions setOrder and setLimit will affect BOTH types
     *
     * @param   string          $table      Name of the table without prefix
     * @param   string|string[] $input      Key => $value array or string with query
     * @param   array           $parameters Optional Parameters for BindValue or to select columns, depends on $input
     * @return  array           SQL Result. Default Data is access able like an obj $player->guid
     */
    public function get($table, $input = array(), $parameters = array()) {
        if (is_array($input)) {
            if (!empty($parameters)) {
                if (is_array($parameters)) {
                    $params = implode(',', $parameters);
                } else {
                    $params = $parameters;
                }
                $sql = 'SELECT ' . $params . ' FROM ' . $this->prefix . $table . ' WHERE 1';
            } else {
                $sql = 'SELECT * FROM ' . $this->prefix . $table . ' WHERE 1';
            }
            foreach ($input as $key => $value) {
                $sql .= ' AND `' . $key . '` = :col_' . $key;
            }
            $sql = rtrim($sql, ',');
            $sql .= $this->order;
            $sql .= $this->limit;
            $this->sth = $this->dbh->prepare($sql);
            foreach ($input as $key => $value) {
                $this->bindValue(':col_' . $key, $value);
            }
        } else if (is_string($input)) {
            $input = str_replace(array("<PFX>", "<TABLE>"), array($this->prefix, $this->prefix . $table), $input);
            $this->sth = $this->dbh->prepare($input);
            foreach ($parameters as $key => $value) {
                $this->bindValue($key, $value);
            }
        }

        $this->sth->execute();
        $this->clearLimit();
        $this->clearOrder();

        if (!is_numeric($this->fetch_type)) {
            $data = $this->sth->fetchAll();
        } else {
            $data = $this->sth->fetchAll($this->fetch_type);
            $this->fetch_type = NULL;
        }
        return $data;
    }

    /**
     * Gets the first result from a database
     *
     * If you pass an array as input you can directly affect the where clause
     * Example:
     * $db->get("mam_users", array("id",5)); => "SELECT * FROM mam_users WHERE id = 5
     *
     * If you only need some fields you can pass a numeric array with the fields you want
     * Example:
     * $db->get("mam_users", array("id",5), array("guid", "lastseen")); => "SELECT guid,lastseen FROM mam_users WHERE id = 5
     *
     * If you pass a string as input, you'll pass a sql query
     * Example:
     * $db->get("mam_users", "SELECT * FROM <TABLE> WHERE guid = :guid, array(":guid" => $guid));
     * For SQL Query it's possible to use the placeholders <TABLE> for table with prefix or <PFX> for prefix
     *
     * The functions setOrder and setLimit will affect BOTH types
     *
     * @param   string          $table      Name of the table without prefix
     * @param   string|string[] $input      Key => $value array or string with query
     * @param   array           $parameters Optional Parameters for BindValue or to select columns, depends on $input
     * @return  array           SQL Result. Default Data is access able like an obj $player->guid
     */
    public function row($table, $input = array(), $parameters = array()) {
        $this->setLimit(0, 1);
        $result = $this->get($table, $input, $parameters);
        if (is_array($result) && !empty($result)) {
            return $result[0];
        } elseif (empty($result)) {
            return false;
        } else {
            return $result;
        }
    }

    /**
     * Upddates a database entry
     *
     * @param string $table Name of the table without prefix
     * @param array $entrys Key Value Array
     * @param array $condition Array with 3 Entrys 1)Column Name 2) Condition 3) Value Example [1] id [2] > [3] 15
     * @return int affected rows
     */
    public function update($table, $entrys, $condition) {
        $sql = 'UPDATE ' . $this->prefix . $table;
        $sql .= ' SET ';
        foreach ($entrys as $key => $value) {
            $sql .= '`' . $key . '` = :col_' . $key . ',';
        }
        $sql = rtrim($sql, ',');
        if (!is_int($condition[2]) && !is_bool($condition[2])) {
            $condition[2] = '"' . $condition[2] . '"';
        }
        $sql .= ' WHERE `' . $condition[0] . '` ' . $condition[1] . ' ' . $condition[2];

        $this->sth = $this->dbh->prepare($sql);
        foreach ($entrys as $key => $value) {
            $this->bindValue(':col_' . $key, $value);
        }
        $this->sth->execute();
        return $this->sth->rowCount();
    }

    /**
     * Insert new data into the database
     *
     * @param string $table Name of the table without prefix
     * @param array $entrys $key => $value where key is the column name
     * @return boolean True if successful else false
     */
    public function insert($table, $entrys) {
        $sql = 'INSERT INTO ' . $this->prefix . $table;
        $sql .= ' (' . implode(',', array_keys($entrys));
        $sql .= ') VALUES (';
        foreach ($entrys as $key => $value) {
            $sql .= ':col_' . $key . ',';
        }
        $sql = rtrim($sql, ',');
        $sql .= ')';
        $this->sth = $this->dbh->prepare($sql);
        foreach ($entrys as $key => $value) {
            $this->bindValue(':col_' . $key, $value);
        }
        if (!$this->sth->execute()) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * Deletes an entry from Database
     *
     * @param string $table Name of the table without prefix
     * @param string $column Name of the column
     * @param mixed $value Value to sarch for in the overgiven column
     * @return int affected rows
     */
    public function delete($table, $column, $value) {
        $sql = 'DELETE FROM ' . $this->prefix . $table;
        $sql .= ' WHERE ' . '`' . $column . '`';
        $sql .= ' = ';
        if (!is_int($value) && !is_bool($value)) {
            $value = '"' . $value . '"';
        }
        $sql .= $value;

        return $this->dbh->exec($sql);
    }

    /**
     * Returns the last inserted id or NULL
     *
     * @return int
     */
    public function getLastInsertedId() {
        return $this->dbh->lastInsertId();
    }

    /**
     * Direclty query the database
     *
     * @param string $sql The sql query
     *
     * @return \PDOStatement The Result
     */
    public function query($sql) {
        $sql = str_replace('<PFX>', $this->prefix, $sql);
        return $this->dbh->query($sql);
    }

    /**
     * Executes a query
     *
     * @param string $sql The sql query
     *
     * @return int Affected Rows
     */
    public function exec($sql) {
        $sql = str_replace('<PFX>', $this->prefix, $sql);
        return $this->dbh->exec($sql);
    }

    /**
     * Prepare SQL Statement
     * @param string $sql The sql query
     */
    public function prepare($sql) {
        $sql = str_replace('<PFX>', $this->prefix, $sql);
        $this->sth = $this->dbh->prepare($sql);
    }

    /**
     * Executes a statement
     *
     * @return \PDOStatement The Result
     */
    public function execute() {
        return $this->sth->execute();
    }

    /**
     * Fetches the Data after a statement
     * @param int $mode
     * @return mixed depends on the mode
     */
    public function fetch($mode) {
        return $this->sth->fetch($mode);
    }

    /**
     * Start a transaction
     *
     * It's possible to execute multiple querys and commit them
     *
     * @return boolean
     */
    public function beginTransaction() {
        return $this->dbh->beginTransaction();
    }

    /**
     * Ends a transaction
     *
     * It's possible to execute multiple querys and commit them
     *
     * @return boolean
     */
    public function endTransaction() {
        return $this->dbh->commit();
    }

    /**
     * Cancel a transaction
     *
     * It's possible to execute multiple querys and commit them
     *
     * @return boolean
     */
    public function cancelTransaction() {
        return $this->dbh->rollBack();
    }

    /**
     * Finds the fitting PDO Typoe for a parameter
     * @param mixed $param string bool or int
     * @return int
     */
    private function getPDOType($param) {
        if (is_string($param)) {
            return PDO::PARAM_STR;
        } elseif (is_int($param)) {
            return PDO::PARAM_INT;
        } elseif (is_bool($param)) {
            return PDO::PARAM_BOOL;
        } else {
            return PDO::PARAM_STR;
        }
    }
}
