<?php
/**
 * ManuAdminMod
 *
 * This is a powerful and platform independent software for administrating game servers of various kinds.
 * If you need help with installing or using this software, please visit our website at: www.ManuAdminMod.de
 * If you want to obtain additional permissions extending the license, contact us at: Webmaster@ManuAdminMod.de
 *
 * @copyright www.ManuAdminMod.de
 * @license http://www.creativecommons.org/licenses/by-nc-nd/4.0/ Creative Commons BY-NC-ND 4.0
 * @version 1.0.0-Beta+7
**/

namespace MAM\Daemon\Libraries\GeoIP;

/**
 * This class opens and decodes a MaxMind GeoIP database. Query geo information
 * with the getLocationData method.
 *
 * @author Jens Forstmann
 * @link http://maxmind.github.io/MaxMind-DB/
 * @link http://dev.maxmind.com/geoip/geoip2/geolite2/
 */
class GeoIP {
    /**
     * Mapping: data field identifier (int) => field type (string)
     * @var string[]
     */
    private $field_types = array(
        0 => 'extended',
        1 => 'pointer',
        2 => 'utf8string',
        3 => 'double',
        4 => 'bytes',
        5 => 'uint16',
        6 => 'uint32',
        7 => 'map',
        8 => 'sint32',
        9 => 'uint64',
        10 => 'uint128',
        11 => 'array',
        12 => 'datacachecontainer',
        13 => 'endmarker',
        14 => 'boolean',
        15 => 'float'
    );
    /**
     * MaxMind-DB metadata
     * @var array
     */
    private $metadata = array();
    /**
     * Start sequence of the metadata part
     * @var string
     */
    private $metadata_start_sequence = "\xab\xcd\xefMaxMind.com";
    /**
     * Size of the search tree in bytes
     * @var int
     */
    private $search_tree_size;
    /**
     * File handle holding the MaxMind-DB file
     * @var resource
     */
    private $file_handle;
    /**
     * Class for Logging
     * @var \MAM\Daemon\Libraries\Logging\LogHandler
     */
    protected $logging;

    /**
     * Initiates a GeoIP object and tries to automatically open a MaxMind-DB.
     * @throws NoDatabaseAvailableException if no suitable MaxMind-DB could be openend
     * @throws GeoIpException
     */
    public function __construct() {
        $this->logging = \MAM\Daemon\Core\Registry::getInstance()->logging;

        // if the first file cannot be opened, it will try the next one and so on...
        $preferred_databases = array('GeoLite2-City.mmdb', 'GeoLite2-Country.mmdb', 'geoip.mmdb');
        $preferred_databases_count = count($preferred_databases);
        for ($i = 0; $i < $preferred_databases_count; $i++) {
            try {
                if ($this->open(__DIR__ . DIRECTORY_SEPARATOR . $preferred_databases[$i])) {
                    break;
                }
            } catch (IOException $ex) {
                // do not mind about the exception, go on
            }
        }
        if (!is_resource($this->file_handle)) {
            $error_message = 'GeoIP: No database available, tried to open: ' . implode(', ', $preferred_databases);
            $this->logging->warning($error_message);
            throw new NoDatabaseAvailableException($error_message);
        }
        $this->logging->info('GeoIP: Initialization finished');
    }

    /**
     * Destructor closes the file.
     */
    public function __destruct() {
        $this->logging->debug('GeoIP: Destructing');
        $this->close();
    }

    /**
     * Returns a string if object is printed
     *
     * Notice: all plugins which where moved to core needs a __tostring function
     * to get it working with the permissions system. Therefore the return of __tostring must me the
     * namespace like deamon, punishment, permissions ...
     *
     * @see http://php.net/manual/en/language.oop5.magic.php#object.tostring
     * @return string
     */
    public function __toString() {
        return __CLASS__ . ' object';
    }

    /**
     * Tries to open and initializes the MaxMind-DB.
     * @param string $filename path to the MaxMind-DB
     * @return boolean true on success, false if file does not exist or is not readable
     * @throws IOException if fopen returns false
     * @throws GeoIpException
     */
    private function open($filename) {
        if (!is_readable($filename)) {
            $this->logging->debug('GeoIP: Database ' . $filename . ' does not exist or is not readable');
            return false;
        }
        // try to open file
        $new_file_handle = fopen($filename, 'rb');
        if ($new_file_handle === false) {
            $error_message = 'GeoIP: Opening the database ' . $filename . ' causes a fopen error';
            $this->logging->warning($error_message);
            throw new IOException($error_message);
        }
        // try to close old file
        $this->close();

        // init some things
        $this->file_handle = $new_file_handle;
        $this->metadata = $this->decode($this->getMetadataStartPosition())[0];
        $this->search_tree_size = (($this->metadata['record_size'] * 2) / 8) * $this->metadata['node_count'];

        $this->logging->info('GeoIP: Database ' . $filename . ' successfully opened');

        return true;
    }

    /**
     * Closes the file.
     * @throws IOException if fclose returns false
     */
    private function close() {
        if (is_resource($this->file_handle)) {
            $this->logging->debug('GeoIP: Closing database');
            if (!fclose($this->file_handle)) {
                $error_message = 'GeoIP: Closing the database causes a fclose error';
                $this->logging->warning($error_message);
                throw new IOException($error_message);
            }
        }
    }

    /**
     * Reads bytes from the MaxMind-DB file.
     * @param int $start_pos position to start at
     * @param int $length count of bytes to read
     * @return string raw byte string
     * @throws IOException if fread returns false or less than $length bytes
     */
    private function readBytes($start_pos, $length) {
        if ($length === 0) {
            return '';
        }
        if (fseek($this->file_handle, $start_pos) !== 0) {
            $error_message = 'GeoIP: Reading bytes from database causes a fseek error';
            $this->logging->warning($error_message);
            throw new IOException($error_message);
        }
        $ret = fread($this->file_handle, $length);
        if (strlen($ret) !== $length) {
            $error_message = 'GeoIP: Reading bytes from database causes a fread error';
            $this->logging->warning($error_message);
            throw new IOException($error_message);
        }
        return $ret;
    }

    /**
     * Fills a string with zeros ("\x00") to get a string with a length of $length.
     * @param string $bytes input data string
     * @param int $length wanted length
     * @param string $fill data to fill the string with
     * @param int $direction STR_PAD_RIGHT|STR_PAD_LEFT|STR_PAD_BOTH
     * @return string padded string
     */
    private function fill($bytes, $length, $fill = "\x00", $direction = STR_PAD_LEFT) {
        return str_pad($bytes, $length, $fill, $direction);
    }

    /**
     * Checks the local byte order and reverts if needed the order to get a big endian format.
     * @param string $bytes input data string
     * @return string big endian byte string
     */
    private function checkByteOrder($bytes) {
        if (unpack('S', "\x01\x00")[1] === 1) {
            // little endian machine, revert byte order
            return strrev($bytes);
        }
        return $bytes;
    }

    /**
     * Searches for the start position of the metadata section.
     * @return int absolut file position of the first byte of the metadata
     * @throws GeoIpException if the metadata start position could not found
     */
    private function getMetadataStartPosition() {
        $file_size = fstat($this->file_handle)['size'];
        $marker_length = strlen($this->metadata_start_sequence);
        for ($pos = $file_size - $marker_length; $pos >= 0; $pos--) {
            if ($this->readBytes($pos, $marker_length) === $this->metadata_start_sequence) {
                $start_position = $pos + $marker_length;
                $this->logging->debug('GeoIP: Found database metadata start position at' . $start_position);
                return $start_position;
            }
        }
        $error_message = 'GeoIP: Could not find the database metadata start position';
        $this->logging->warning($error_message);
        throw new GeoIpException($error_message);
    }

    /**
     * Calculates the size for a data field
     * @param string $control_byte
     * @param int $position position of the next byte (first byte after the control byte)
     * @return array 0=>size, 1=>next position
     */
    private function getSizeFromControlByte($control_byte, $position) {
        $size = $control_byte & 0b11111; // [0-31]
        $read_additional_bytes = $size > 28 ? $size & 0b11 : 0; // [0-3]
        list($to_add, $position) = $this->decodeUint16or24or32($position, $read_additional_bytes);

        if ($size === 29) {
            $size = 29 + $to_add;
        } else if ($size === 30) {
            $size = 285 + $to_add;
        } else if ($size === 31) {
            $size = 65821 + $to_add;
        }
        return array($size, $position);
    }

    /**
     * Searches for the location data for a given IPv4 or IPv6 address.
     * @param string $ip_address IPv4 or IPv6 address
     * @return array|boolean false if no data exists for this ip address, otherwise the location data
     * @throws InvalidIpAddressException if $ip_address is invalid
     * @throws GeoIpException
     */
    public function getLocationData($ip_address) {
        $this->logging->debug('GeoIP: Get location data for ip address ' . $ip_address);
        if (!filter_var($ip_address, FILTER_VALIDATE_IP)) {
            $error_message = 'GeoIP: ' . $ip_address . ' is an invalid ip address';
            $this->logging->warning($error_message);
            throw new InvalidIpAddressException($error_message);
        }

        $ip_address_array = array_merge(unpack('C*', $this->fill(inet_pton($ip_address), 128 / 8))); // always 128 bit, each byte in one array slot

        $current_node = 0;
        for ($i = 0; $i < 128; $i++) {
            if ($current_node === $this->metadata['node_count']) {
                $this->logging->info('GeoIP: No location data found for ip address ' . $ip_address);
                return false; // no data for this ip address
            }
            if ($current_node > $this->metadata['node_count']) {
                // data available
                $position = $current_node - $this->metadata['node_count'] + $this->search_tree_size;
                $this->logging->debug('GeoIP: Location data for ip address ' . $ip_address . ' found at position ' . $position);
                return $this->decode($position)[0];
            }
            $current_bit = 1 & ($ip_address_array[$i / 8] >> (7 - $i % 8));
            $current_node = $this->decodeNode($current_node, $current_bit);
        }

        // if we reach this => buggy implementation
        $error_message = 'GeoIP: getLocationData error for ip address ' . $ip_address;
        $this->logging->warning($error_message);
        throw new GeoIpException($error_message);
    }

    /**
     * Decodes a field.
     * @param int $position position of the field to decode, pointing at the control byte
     * @return array 0=>field data, 1=>next position
     * @throws GeoIpException
     */
    private function decode($position) {
        $control_byte = unpack('C', $this->readBytes($position, 1))[1];
        $position++;

        $field_type = $this->field_types[$control_byte >> 5];

        if ($field_type === 'extended') {
            $next_byte = unpack('C', $this->readBytes($position, 1))[1];
            $position++;
            if ($next_byte + 7 <= 7) {
                $error_message = 'GeoIP: Field decoding causes an error with extended fields';
                $this->logging->warning($error_message);
                throw new GeoIpException($error_message);
            }
            $field_type = $this->field_types[$next_byte + 7];
        }
        $old_position = $position;
        list($size, $position) = $this->getSizeFromControlByte($control_byte, $position);
        switch ($field_type) {
            case 'map':
                return $this->decodeMap($position, $size);
            case 'utf8string':
                return $this->decodeUTF8String($position, $size);
            case 'uint16':
            case 'uint32':
                return $this->decodeUint16or24or32($position, $size);
            case 'uint64':
            case 'uint128':
                return $this->decodeUint64or128($position, $size);
            case 'array':
                return $this->decodeArray($position, $size);
            case 'pointer':
                return $this->decodePointer($old_position, $control_byte);
            case 'sint32':
                return $this->decodeSint32($position, $size);
            case 'double':
                return $this->decodeDouble($position, $size);
            case 'float':
                return $this->decodeFloat($position, $size);
            default:
                // TODO implement decode methods for all field types
                $error_message = 'GeoIP: No implementation for decoding field type ' . $field_type;
                $this->logging->warning($error_message);
                throw new GeoIpException($error_message);
        }
    }

    /**
     * Decodes an array field.
     * @param int $position position of the field, pointing to the first data byte
     * @param int $size number of array elements
     * @return array 0=>zero indexed array, 1=>next position
     */
    private function decodeArray($position, $size) {
        $array = array();

        for ($i = 0; $i < $size; $i++) {
            list($value, $position) = $this->decode($position);
            $array[] = $value;
        }

        return array($array, $position);
    }

    /**
     * Decodes a map field.
     * @param int $position position of the field, pointing to the first data byte
     * @param int $size number of map elements
     * @return array 0=>associative array, 1=>next position
     */
    private function decodeMap($position, $size) {
        $map = array();

        for ($i = 0; $i < $size; $i++) {
            list($key, $position) = $this->decode($position);
            list($value, $position) = $this->decode($position);
            $map[$key] = $value;
        }

        return array($map, $position);
    }

    /**
     * Decodes a pointer. That means to calculate the address and decodes that field.
     * @param int $position pointing at the field after the control byte
     * @param string $control_byte
     * @return array 0=>field data, 1=>next position (field after the pointer)
     */
    private function decodePointer($position, $control_byte) {
        $size = (($control_byte & 0b11000) >> 3) + 1;
        $value = $control_byte & 0b111;
        $bytes = $this->readBytes($position, $size);

        if ($size !== 4) {
            $bytes = pack('C', $value) . $bytes;
        }

        $pointer_offsets = array(
            1 => 0,
            2 => 2048,
            3 => 526336,
            4 => 0,
        );

        $pointer = unpack('N', $this->fill($bytes, 4))[1] + $pointer_offsets[$size] + $this->search_tree_size + 16;

        return array($this->decode($pointer)[0], $position + $size);
    }

    /**
     * Decodes a node of the search tree containing two records.
     * Currently only supports record sizes of 24, 28 and 32 length.
     * @param int $node_number number of the node to decode
     * @param int $bit current bit of the search (0|1)
     * @return int next node number
     * @throws GeoIpException if record_size is not supported
     */
    private function decodeNode($node_number, $bit) {
        $supported_record_sizes = array(24, 28, 32); // for more than 32 bits, code for big int is needed!
        if (!in_array($this->metadata['record_size'], $supported_record_sizes, true)) {
            $error_message = 'GeoIP: The current database uses a record size of '
                           . $this->metadata['record_size']
                           . ', but currently only the following sizes are supported: '
                           . implode(', ', $supported_record_sizes);
            $this->logging->warning($error_message);
            throw new GeoIpException($error_message);
        }
        $node_size = $this->metadata['record_size'] / 4;
        $position = $node_number * $node_size;
        $bytes_of_node = $this->readBytes($position, $node_size);
        $bytes = substr($bytes_of_node, $bit * ceil($node_size / 2), ceil(($this->metadata['record_size'] - 4) / 8));
        if ($this->metadata['record_size'] % 8 !== 0) {
            // 28, 36, ... (splitted byte in the middle)
            $middle = unpack('C', substr($bytes_of_node, $node_size / 2, 1))[1]; // need integer for bitwise shifting
            if ($bit === 1) {
                $middle = $middle & 0x0F;
            } else {
                $middle = (0xF0 & $middle) >> 4;
            }
            $bytes = chr($middle) . $bytes;
        }

        return unpack('N', $this->fill($bytes, 4))[1];
    }

    /**
     * Decodes an UTF8 string.
     * @param int $position position of the field, pointing to the first data byte
     * @param int $size string length in bytes
     * @return array 0=>string, 1=>next position
     */
    private function decodeUTF8String($position, $size) {
        $string = $this->readBytes($position, $size);
        return array($string, $position + $size);
    }

    /**
     * Decodes a floating point number.
     * @param int $position position of the field, pointing to the first data byte
     * @param int $size number of bytes used for the float representation at $position, must be 4
     * @return array 0=>float, 1=>next position
     * @throws GeoIpException if $size is not 4
     */
    private function decodeFloat($position, $size) {
        if ($size !== 4) {
            $error_message = 'GeoIP: Decoding a float is not possible for size ' . $size;
            $this->logging->warning($error_message);
            throw new GeoIpException($error_message);
        }
        $bytes = $this->readBytes($position, $size);
        return array(unpack('f', $this->checkByteOrder($bytes))[1], $position + $size);
    }

    /**
     * Decodes a double point number.
     * @param int $position position of the field, pointing to the first data byte
     * @param int $size number of bytes used for the double representation at $position, must be 8
     * @return array 0=>double, 1=>next position
     * @throws GeoIpException if $size is not 8
     */
    private function decodeDouble($position, $size) {
        if ($size !== 8) {
            $error_message = 'GeoIP: Decoding a double is not possible for size ' . $size;
            $this->logging->warning($error_message);
            throw new GeoIpException($error_message);
        }
        $bytes = $this->readBytes($position, $size);
        return array(unpack('d', $this->checkByteOrder($bytes))[1], $position + $size);
    }

    /**
     * Decodes a signed 32bit integer.
     * @param int $position position of the field, pointing to the first data byte
     * @param int $size number of bytes used for the signed integer representation at $position
     * @return array 0=>signed int, 1=>next position
     */
    private function decodeSint32($position, $size) {
        $bytes = $this->fill($this->readBytes($position, $size), 4);
        return array(unpack('l', $this->checkByteOrder($bytes))[1], $position + $size);
    }

    /**
     * Decodes an unsigned 16/24/32 bit integer.
     * @param int $position position of the field, pointing to the first data byte
     * @param int $size number of bytes used for the unsigned integer representation at $position
     * @return array 0=>unsigned int, 1=>next position
     */
    private function decodeUint16or24or32($position, $size) {
        $bytes = $this->fill($this->readBytes($position, $size), 4);
        return array(unpack('N', $bytes)[1], $position + $size);
    }

    /**
     * Decodes an unsigned 64/128 bit integer.
     * @param int $position position of the field, pointing to the first data byte
     * @param int $size number of bytes used for the unsigned integer representation at $position
     * @return array 0=>unsigned int, 1=>next position
     * @throws GeoIpException if $size > 4, this is currently not implemented
     */
    private function decodeUint64or128($position, $size) {
        if ($size > 4) {
            $error_message = 'GeoIP: Decoding an unsigned big int is not implemented for size ' . $size;
            $this->logging->warning($error_message);
            throw new GeoIpException($error_message);
        }
        return $this->decodeUint16or24or32($position, $size);
    }
}
