<?php
/**
 * ManuAdminMod
 *
 * This is a powerful and platform independent software for administrating game servers of various kinds.
 * If you need help with installing or using this software, please visit our website at: www.ManuAdminMod.de
 * If you want to obtain additional permissions extending the license, contact us at: Webmaster@ManuAdminMod.de
 *
 * @copyright www.ManuAdminMod.de
 * @license Creative Commons BY-NC-ND 4.0 <http://www.creativecommons.org/licenses/by-nc-nd/4.0/>
 * @version 1.0.0-Beta+7
 * */

namespace MAM\Daemon\Libraries\Permissions;

use MAM\Daemon\Core\Utils;

/**
 * Group Commands
 *
 * This class implements all command about permissions
 * ->set to set a player to a specific group
 * ->check <PLAYER> to get the group of a player
 * ->check to get your own group
 * ->createtoken <GROUP> to create token for a group
 * ->usetoken <TOKEN> to use the token
 *
 * @author Mirko911 <mirko911@manuadminmod.de>
 * @version 1.0.0
 */
class Commands extends Utils {

    /**
     * Returns a string if object is printed
     *
     * Notice: all plugins which where moved to core needs a __tostring function
     * to get it working with the permissions system. Therefore the return of __tostring must me the
     * namespace like deamon, punishment, permissions ...
     *
     * @see http://php.net/manual/en/language.oop5.magic.php#object.tostring
     * @return string
     */
    public function __toString() {
        return 'permissions';
    }

    /**
     * Constructor
     */
    public function __construct() {
        parent::__construct();
        $this->runPlugin();
    }

    /**
     * Runs the plugin
     *
     * This function is called after all plugins were loaded and dependencies were checked.
     * In this function you can define things which you want to load, when all plugins are fine.
     *
     * If you know that you don't have any dependencies, you can remove this function.
     * Otherwise you should use it.
     *
     * In this function you should call functions like registerCommand, registerEvent, triggerEvent
     */
    public function runPlugin() {
        parent::runPlugin();

        $this->mod->registerCommand('group', '~^group \S.*$~i', 'groupCommand', $this);
        $stored_tokens = \MAM\Daemon\Libraries\Helper\Helper::parseJson($this->mod->getConfigDir() . 'storage/tokens.json', false);

        if (count($this->mod->getAdmins()) == 0 && empty($stored_tokens)) {
            $token = $this->addToken('ops');
            $text = array();
            $text[] = 'CREATED ADMIN TOKEN';
            $text[] = $token;
            $text[] = 'You can use it ingame over ';
            $text[] = '!group usetoken ' . $token;
            \MAM\Daemon\Libraries\Helper\Helper::createInfoBox($text, '*', 'info');
        }
    }

    /**
     * Splits the overgiven command in sub commands
     *
     * This functions checks the parameters and calls the fitting function for the overgiven parameter
     *
     * @param   string      $guid       Guid of executing player
     * @param   string[]    $parameters The chatline splitted by " " without !command
     */
    public function groupCommand($guid, $parameters) {
        $command = strtolower(array_shift($parameters));
        switch ($command) {
            case 'set':
                $this->setGroup($guid, $parameters);
                break;
            case 'check':
                if (empty($parameters)) {
                    $this->getOwnGroup($guid, $parameters);
                } else {
                    $this->getGroup($guid, $parameters);
                }
                break;
            case 'createtoken':
                $this->createToken($guid, $parameters);
                break;
            case 'usetoken':
                $this->useToken($guid, $parameters);
                break;
            default :
                $this->mod->printHelp('group', $guid);
                break;
        }
    }

    /**
     * Adds a player to a group
     *
     * @param   string      $guid       Guid of executing player
     * @param   string[]    $parameters The chatline splitted by " " without !command
     * @return  false       false on fail
     */
    private function setGroup($guid, $parameters) {
        if (empty($parameters)) {
            $this->mod->printHelp('setgroup', $guid);
            return false;
        }
        $parameters = array_map('strtolower', $parameters);
        $group = array_shift($parameters);
        $player = array_shift($parameters);

        if (!array_key_exists($group, $this->mod->getGroups())) {
            $list = implode(', ', array_keys($this->mod->getGroups()));
            $this->players[$guid]->say($this->mod->getLngString('availableGroups', array('<GROUPS>'), array($list)));
            return false;
        }

        $tochange = $this->mod->findPlayerGuid($player);

        if (!$tochange) {
            $this->players[$guid]->say($this->mod->getLngString('playerNotFound', array('<SEARCH>'), array($player)));
            return false;
        }

        if (!$this->permissions->hasPermissions($guid, 'permissions.group.set.' . $group)) {
            $this->players[$guid]->say($this->mod->getLngString('morePowerNeeded'));
            return false;
        }

        $this->players[$tochange]->setGroup($group);

        $this->rcon->rconSay($this->mod->getLngString('groupChanged', array('<GROUP>', '<PLAYER_NAME>'), array($this->mod->getLongGroupName($group), $this->players[$tochange]->getName())));
    }

    /**
     * Checks if a admin is in a admin group
     *
     * @param   string      $guid       Guid of executing player
     * @param   string[]    $parameters The chatline splitted by " " without !command
     * @return boolean
     */
    public function getGroup($guid, $parameters) {
        $parameters = implode(' ', $parameters);

        $player = $this->mod->findPlayerGuid($parameters);
        if (!$player) {
            $this->players[$guid]->say($this->mod->getLngString('playerNotFound', array('<SEARCH>'), array($parameters)));
            return false;
        }
        $group = $this->players[$player]->getGroup();
        if (!$this->permissions->hasPermissions($guid, 'permissions.group.check.' . $group)) {
            $this->players[$guid]->say($this->mod->getLngString('morePowerNeeded'));
            return false;
        }
        $group = $this->mod->getLongGroupName($group);
        $name = $this->players[$player]->getName();

        $this->rcon->rconSay($this->mod->getLngString('checkAdmin', array('<PLAYER_GROUP>', '<PLAYER_NAME>'), array($group, $name)));
    }

    /**
     * Tests in which group you are
     *
     * @param   string      $guid       Guid of executing player
     * @param   string[]    $parameters The chatline splitted by " " without !command
     */
    public function getOwnGroup($guid, $parameters) {
        if (!$this->permissions->hasPermissions($guid, 'permissions.group.checkowngroup')) {
            $this->players[$guid]->say($this->mod->getLngString('morePowerNeeded'));
            return false;
        }
        $nick = $this->players[$guid]->getName();
        $group = $this->mod->getLongGroupName($this->players[$guid]->getGroup());
        $this->rcon->rconSay($this->mod->getLngString('admintest', array('<PLAYER_NAME>', '<PLAYER_GROUP>'), array($nick, $group)));
    }

    /**
     * Command for creating admin tokens
     *
     * With this command you can create a one time admin token for a group you choose
     *
     * @param   string      $guid       Guid of executing player
     * @param   string[]    $parameters The chatline splitted by " " without !command
     * @return boolean false on fail
     */
    public function createToken($guid, $parameters) {
        if (empty($parameters)) {
            $this->mod->printHelp('createtoken', $guid);
            return;
        }
        $group = implode(' ', array_map('strtolower', $parameters));

        if (!array_key_exists($group, $this->mod->getGroups())) {
            $list = implode(', ', array_keys($this->mod->getGroups()));
            $this->players[$guid]->say($this->mod->getLngString('availableGroups', array('<GROUPS>'), array($list)));
            return false;
        }

        if (!$this->permissions->hasPermissions($guid, 'permissions.group.createtoken' . $group)) {
            $this->players[$guid]->say($this->mod->getLngString('morePowerNeeded'));
            return false;
        }

        $token = $this->addToken($group);
        $this->players[$guid]->say($this->mod->getLngString('createdToken', array('<TOKEN>', '<GROUP>'), array($token, $group)));
    }

    /**
     * Command for using an admin token
     *
     * This command checks if the token exists in tokens list.
     * If the token exists, the group of the player changes
     * and the token will be removed from json file
     *
     * @param   string      $guid       Guid of executing player
     * @param   string[]    $parameters The chatline splitted by " " without !command
     * @return boolean false on fail
     */
    public function useToken($guid, $parameters) {
        if (empty($parameters)) {
            $this->mod->printHelp('usetoken', $guid);
            return;
        }
        if (!$this->permissions->hasPermissions($guid, 'permissions.group.usetoken')) {
            $this->players[$guid]->say($this->mod->getLngString('morePowerNeeded'));
            return false;
        }
        $parameters = implode(' ', $parameters);
        $stored_tokens = \MAM\Daemon\Libraries\Helper\Helper::parseJson($this->mod->getConfigDir() . 'storage/tokens.json', false);
        if (array_key_exists($parameters, $stored_tokens)) {
            $this->players[$guid]->setGroup($stored_tokens[$parameters]);
            $this->players[$guid]->say($this->mod->getLngString('usedToken'));

            unset($stored_tokens[$parameters]);
            if (empty($stored_tokens)) {
                $json = json_encode(json_decode('{}'), JSON_PRETTY_PRINT | JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE);
            } else {
                $json = json_encode($stored_tokens, JSON_PRETTY_PRINT | JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE);
            }
            file_put_contents($this->mod->getConfigDir() . 'storage/tokens.json', $json);
        } else {
            $this->players[$guid]->say($this->mod->getLngString('anyError'));
        }
    }

    /**
     * Function to create tokens
     *
     * This function generates a random 8 char long token.
     * The created token IS case-sensitive
     *
     * @param string $group Name of the group you want to create the token for
     * @return string 8 char long Token
     */
    public function addToken($group) {
        $stored_tokens = \MAM\Daemon\Libraries\Helper\Helper::parseJson($this->mod->getConfigDir() . 'storage/tokens.json', false);

        $alphabet = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890+-*/-_.,:()!?';
        $token = '';
        for ($i = 0; $i < 8; $i++) {
            $token .= $alphabet[rand(0, strlen($alphabet) - 1)];
        }

        $stored_tokens[$token] = $group;
        file_put_contents($this->mod->getConfigDir() . 'storage/tokens.json', json_encode($stored_tokens, JSON_PRETTY_PRINT | JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE));
        return $token;
    }
}
