<?php
/**
 * ManuAdminMod
 *
 * This is a powerful and platform independent software for administrating game servers of various kinds.
 * If you need help with installing or using this software, please visit our website at: www.ManuAdminMod.de
 * If you want to obtain additional permissions extending the license, contact us at: Webmaster@ManuAdminMod.de
 *
 * @copyright www.ManuAdminMod.de
 * @license http://www.creativecommons.org/licenses/by-nc-nd/4.0/ Creative Commons BY-NC-ND 4.0
 * @version 1.0.0-Beta+7
 * */

namespace MAM\Daemon\Libraries\Plugins;

use MAM\Daemon\Core\Registry;
use MAM\Daemon\Libraries\Helper\Helper;

/**
 * This Class stores all information about a plugin
 *
 * @author Mirko911 <mirko911@manuadminmod.de>
 * @version 1.0.0
 */
class PluginLoader {
    /**
     * Array of all loaded plugins
     *
     * @var array all loaded plugins
     */
    private $plugins = array();
    /**
     * Class for mam related stuff
     *
     * @var \MAM\Daemon\Core\Mod
     */
    private $mod;
    /**
     * Loghandler
     *
     * @var \MAM\Daemon\Libraries\Logging\Loghandler
     */
    private $logging;

    /**
     * Constructor
     */
    public function __construct() {
        $registry = Registry::getInstance();
        $this->mod = $registry->mod;
        $this->logging = $registry->logging;
    }

    /**
     * Returns a string if object is printed
     *
     * Notice: all plugins which where moved to core needs a __tostring function
     * to get it working with the permissions system. Therefore the return of __tostring must me the
     * namespace like deamon, punishment, permissions ...
     *
     * @see http://php.net/manual/en/language.oop5.magic.php#object.tostring
     * @return string
     */
    public function __toString() {
        return 'PLuginLoader Class';
    }

    /**
     * Loads all plugins from a specific dir
     *
     * @param string $dir Path to dir
     * @return int Number of Plugins loaded
     */
    public function loadPluginDir($dir) {


        $dirfp = opendir($dir . DIRECTORY_SEPARATOR);
        $i = 0;
        while (($file = readdir($dirfp)) !== false) {
            if (in_array($file, array('.', '..'))) {
                continue;
            }

            if (is_file($dir . DIRECTORY_SEPARATOR . $file . DIRECTORY_SEPARATOR . 'info.json')) {
                $data = Helper::parseJson($dir . DIRECTORY_SEPARATOR . $file . DIRECTORY_SEPARATOR . 'info.json');
            } else {
                $this->logging->warning('Plugin ' . $file . ' skipped. No info.json found');
                continue;
            }
            $namespace = $class_name = key($data);

            $game = $this->mod->getGame();
            if (!isset($data[$namespace]['compatibility']['games'][$game]) || $data[$namespace]['compatibility']['games'][$game] == '0') {
                continue;
            }

            $parts = array_map('ucfirst', explode('.', $class_name));
            $plugin_name = $parts[count($parts) - 1];
            $parts[] = $plugin_name;
            $class_name = implode('\\', $parts);

            $plugin = $data[$namespace];
            $plugin['class'] = new $class_name();

            if (!is_object($plugin['class'])) {
                $this->logging->warning('Plugin ' . $file . ' not loaded. It\'s no class');
                return false;
            }

            $this->plugins[$namespace] = $plugin;

            $this->logging->notice(' + Loading Plugin \'' . $data[$namespace]['information']['name'] . '\' (v' . $data[$namespace]['information']['version'] . ') by ' . $data[$namespace]['information']['author']);
            $i++;
        }

        closedir($dirfp);

        foreach ($this->plugins as $namespace => $data) {
            $this->plugins[$namespace]['class']->initPlugin();
        }

        return $i;
    }

    /**
     * Checks if a plugin has dependencies
     *
     * If MAM can't find the dependency, the plugin gets unloaded
     */
    public function checkDependencies() {
        foreach ($this->plugins as $namespace => $plugin) {
            if (!array_key_exists('dependencies', $plugin['compatibility'])) {
                continue;
            }
            foreach ($plugin['compatibility']['dependencies'] as $key => $dependency) {
                if (!array_key_exists($key, $this->plugins)) {
                    $this->logging->notice(' - Plugin \'' . $this->plugins[$namespace]['information']['name'] . '\' (v' . $this->plugins[$namespace]['information']['version'] . ') has missing dependencies (' . $key . ')');
                    unset($this->plugins[$namespace]);
                    break;
                }
            }
        }
    }

    /**
     * Execute function after depency check
     */
    public function runPlugins() {
        foreach ($this->plugins as $plugin) {
            if (!array_key_exists('disabled', $plugin) || !$plugin['disabled']) {
                $plugin['class']->runPlugin();
            }
        }
    }

    /**
     * Returns a namespace array of all loaded plugins
     *
     * @return array List of all namespaces
     */
    public function getPlugins() {
        return array_keys($this->plugins);
    }

    /**
     * Returns all information about a plugin.
     * The return value contains all data from info.json file + a reference to the created instance of the plugin
     *
     * @param string $namespace The namespace of the plugin
     * @return mixed array if plugin was found, false if not
     */
    public function getPlugin($namespace) {
        if (array_key_exists($namespace, $this->plugins)) {
            return $this->plugins[$namespace];
        } else {
            $this->logging->warning('Plugin ' . $namespace . ' doesn\'t exist or wasn\'t loaded');
            return false;
        }
    }

    /**
     * Returns all plugin meta data
     *
     * Meta data are namespace, name, version, author, description, date, website
     *
     * @param string $namespace The namespace of the plugin
     * @return boolean|array false if plugin doesn't exist, else array of meta data
     */
    public function getPluginData($namespace) {
        if (array_key_exists($namespace, $this->plugins)) {
            $temp_plugin = $this->plugins[$namespace];
            return $temp_plugin['information'];
        } else {
            return false;
        }
    }

    /**
     * Disables a Plugin
     *
     * Disable means, that the object is still there, but there are no registerEvents/registerCommands called.
     * So you can call them later if you want that. Only defaultCvars are registed
     *
     * @param string $namespace The namespace of the plugin . mam.plugins.example You can find the namespace in the info,json of every plugin
     */
    public function disablePlugin($namespace) {
        if (array_key_exists($namespace, $this->plugins)) {
            $this->logging->notice(' - Plugin \'' . $this->plugins[$namespace]['information']['name'] . '\' (v' . $this->plugins[$namespace]['information']['version'] . ') is disabled');
            $this->plugins[$namespace]['disabled'] = true;
        }
    }

    /**
     * Checks if a plugin isset
     *
     * @param string $namespace The namespace of the plugin . mam.plugins.example You can find the namespace in the info,json of every plugin
     * @return boolean true if isset| false if not
     */
    public function issetPlugin($namespace) {
        return (array_key_exists($namespace, $this->plugins)) ? true : false;
    }
}
