<?php
/**
 * ManuAdminMod
 *
 * This is a powerful and platform independent software for administrating game servers of various kinds.
 * If you need help with installing or using this software, please visit our website at: www.ManuAdminMod.de
 * If you want to obtain additional permissions extending the license, contact us at: Webmaster@ManuAdminMod.de
 *
 * @copyright www.ManuAdminMod.de
 * @license http://www.creativecommons.org/licenses/by-nc-nd/4.0/ Creative Commons BY-NC-ND 4.0
 * @version 1.0.0-Beta+7
 * */

namespace MAM\Daemon\Libraries\TwoFactorAuth;

use MAM\Daemon\Core\Utils;

/**
 * Class for handling 2 factor authentications
 * @author  belst <belstgut@manuadminmod.de>
 * @version 1.0.0
 */
class TwoFactorAuth extends Utils {
    /**
     * Totp object. Used for creating and checking tokens
     * @var Totp
     */
    private $totp = null;

    /**
     * Calls Parent constructor and initiates Totp object
     */
    public function __construct() {
        parent::__construct();

        $this->totp = new Totp();

        $this->initPlugin();
        $this->runPlugin();
    }

    /**
     * Initializes the Object and sets Default config values
     * @return void
     */
    public function initPlugin() {
        parent::initPlugin();

        $this->mod->setDefaultCV('twofactorauth', 'enabled', false);
        $this->mod->setDefaultCV('twofactorauth', 'secretlength', 16);

        // initialize already connected users.
        foreach ($this->players as &$player) {
            $player->authed = false;
        }
    }

    /**
     * Runs the Library and registers commands, events and hooks
     * @return void
     */
    public function runPlugin() {
        parent::runPlugin();

        $this->mod->registerCommand('auth', '~^auth \d{6}$~i', 'authCommand', $this);
        $this->mod->registerCommand('generatesecrets', false, 'generateSecretsIfNotExists', $this);
        $this->mod->registerCommand('showsecret', '~^showsecret (NAME .+|GUID [a-fA-F0-9]{8}|PID \d+)$~i', 'showSecret', $this);

        $this->mod->registerEvent('playerJoined', 'falseAuthed', $this);
        $this->mod->registerEvent('adminAdded', 'createSecret', $this);

        $this->mod->registerHook('preCommandExecute', 'checkAuthed', $this);
    }

    /**
     * Authenticates the caller if token is correct
     * @param   string      $guid       Guid of executing player
     * @param   string[]    $parameters The chatline splitted by " " without !command
     * @return  boolean     returns auth status
     */
    public function authCommand($guid, $parameters) {
        $user = $this->players[$guid];
        if (!$this->mod->getCV('twofactorauth', 'enabled')) {
            $user->say($this->mod->getLngString('twoFactorDisabled'));
            return true;
        }

        if ($this->isAuthed($guid)) {
            $user->say($this->mod->getLngString('alreadyAuthed'));
            return true;
        }
        if ($user->getGroup() === 'default') {
            $user->say($this->mod->getLngString('dontAuth4Default'));
            return true;
        }
        if (!isset($this->mod->getAdmins()[$user->getGuid()]['secret'])) {
            $user->say($this->mod->getLngString('noSecretFound'));
            $this->logging->warning($user . ' tried to authenticate, but no secret token was found.');
            return false;
        }

        if ($this->totp->checkCode($this->mod->getAdmins()[$user->getGuid()]['secret'], trim($parameters[0]))) {
            $user->authed = true;
            $user->say($this->mod->getLngString('authSuccess'));
            $this->logging->notice($user . ' authenticated successfully.');
            return true;
        } else {
            $user->say($this->mod->getLngString('tokenInvalid'));
            return false;
        }
    }

    /**
     * Generates A secret token for all admins who lack one
     * @param   string      $guid       Guid of executing player
     * @param   string[]    $parameters The chatline splitted by " " without !command
     * @return  void
     */
    public function generateSecretsIfNotExists($guid, $parameters = false) {
        $admins = $this->mod->getAdmins();
        $num = 0;
        foreach ($admins as $adminguid => $admin) {
            if (!isset($admin['secret'])) {
                $admins[$adminguid]['secret'] = $this->totp->createSecret($this->mod->getCV('twofactorauth', 'secretlength'));
                $num++;
            }
        }

        $this->mod->setAdmins($admins);

        $this->players[$guid]->say($this->mod->getLngString('secretsGenerated', ['<VALUE>', '<PREFIX>'], [$num, $this->mod->getCV('daemon', 'cmdprefix')]));
        $this->logging->notice($this->players[$guid] . ' created secrets for all missing admins. ' . $num . ' admins affected.');
    }

    /**
     * Hooks into executeCommand and prevents execution if player is not authed
     * @param  string $guid    Guid of the calling player
     * @param  string $command the command which gets executed
     * @return boolean         whether a player can execute the given command or not based on his auth status
     */
    public function checkAuthed($guid, $command) {
        if (!$this->mod->getCV('twofactorauth', 'enabled')) {
            return true;
        }

        if ($this->players[$guid]->authed) {
            return true;
        }

        if (in_array($command, $this->mod->getGroupCommands('default'))) {
            return true;
        }

        $this->players[$guid]->say($this->mod->getLngString('missingAuth', ['<PREFIX>'], [$this->mod->getCV('daemon', 'cmdprefix')]));
        return false;
    }

    /**
     * Initializes newly connected players with authed attribute
     * @param  string $guid Guid of the newly connected player
     * @return void
     */
    public function falseAuthed($guid) {
        $this->players[$guid]->authed = false;
    }

    /**
     * Finds player by name or pid if he is only and displays the secret of the player to the caller
     * @param  string $guid       Guid of the caller
     * @param  string $parameters Parameters. has to match <pre>^(NAME .+|GUID [a-fA-F0-9]{8}|PID \d+)$</pre>
     * @see    showByNameOrPid()
     * @see    showByGuid()
     * @return void
     */
    public function showSecret($guid, $parameters) {
        $type = $parameters[0];
        switch (strtolower($type)) {
            case 'name':
            case 'pid':
                $this->showByNameOrPid($guid, $parameters[1]);
                break;
            case 'guid':
                $this->showByGuid($guid, $parameters[1]);
                break;
            default:
                $this->players[$guid]->say($this->mod->getLngString('showSecretInvalid', ['<PREFIX>'], [$this->mod->getCV('daemon', 'cmdprefix')]));
                break;
        }
    }

    /**
     * generates a Secret and adds it to the admins array
     * @param  array  $params contains guid and the adminarray
     * @return void
     */
    public function createSecret(&$params) {
        $guid = $params[0];
        $admins = &$params[1];

        $admins[$guid]['secret'] = $this->totp->createSecret($this->mod->getCV('twofactorauth', 'secretlength'));

        $this->players[$guid]->say($this->mod->getLngString('announceSecret', ['<VALUE>'], [$admins[$guid]['secret']]));
    }

    /**
     * Shows the secret of an online Player identified by name or pid
     * @param  string $guid   Guid of the caller
     * @param  string $search PID or part of nick of the player
     * @return void
     */
    private function showByNameOrPid($guid, $search) {
        $player = $this->mod->findPlayerGuid($search);
        if (!$player) {
            $this->players[$guid]->say($this->mod->getLngString('playerNotFound', array("<SEARCH>"), array($search)));
            return;
        }
        $player = $this->players[$player];
        if ($player->getGroup() === 'default') {
            $this->players[$guid]->say($this->mod->getLngString('noSecret4Default'));
            return;
        }

        if (!isset($this->mod->getAdmins()[$player->getGuid()]['secret'])) {
            $this->players[$guid]->say($this->mod->getLngString('playerMissingSecret', ['<PLAYER>', '<PREFIX>'], [$player, $this->mod->getCV('daemon', 'cmdprefix')]));
            return;
        }

        $this->players[$guid]->say($this->mod->getLngString('showSecret', ['<PLAYER>', '<VALUE>'], [$player, $this->mod->getAdmins()[$player->getGuid()]['secret']]));
    }

    /**
     * Finds a player based on the short 8 character guid and sends his secret to the caller
     * @param  string $guid  Guid of the caller
     * @param  string $param short guid of the admin
     * @return void
     */
    private function showByGuid($guid, $param) {
        $admins = $this->mod->getAdmins();
        $shortadmins = array();
        foreach ($admins as $guid => $admin) {
            $shortadmins[substr($guid, -8)] = $guid;
        }

        if (!isset($shortadmins[$param])) {
            $this->players[$guid]->say($this->mod->getLngString('noAdminFound', ['<VALUE>'], [$param]));
            return;
        }
        $admin = $admins[$shortadmins[$param]];

        if (!isset($admin['secret'])) {
            $this->players[$guid]->say($this->mod->getLngString('noSecret4Player', ['<PLAYER>', '<PREFIX>'], [$admin['names'][0] . '(' . $shortadmins[$param] . ')', $this->mod->getCV('daemon', 'cmdprefix')]));
            return;
        }

        $this->players[$guid]->say($this->mod->getLngString('showSecret', ['<PLAYER>', '<VALUE>'], [$admin['names'][0] . '(' . $shortadmins[$param] . ')', $admin['secret']]));
    }

    /**
     * Returns auth status of a player
     * @param  string  $guid Guid of the player to check
     * @return boolean       auth status or false if undefined
     */
    private function isAuthed($guid) {
        if (!isset($this->players[$guid]->authed)) {
            return false;
        }
        return $this->players[$guid]->authed;
    }
}
